import base64
import hashlib
import socket


def get_internal_ip():
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        s.connect(("8.8.8.8", 80))
        ip = s.getsockname()[0]
    return ip


def get_unique_id(text: str) -> str:
    return "_" + (
        base64.b64encode(hashlib.md5(text.encode()).digest())
        .decode()
        .replace("=", "")
        .replace("/", "_")
        .replace("+", "")
    )



import abc
import time
import logging
from typing import Optional, Dict, Any, Union, Iterable
from pydantic import BaseModel, Field
from mqtt_rest.configs import SERVER_CONFIG as CONFIG
from mqtt_rest.utils import get_unique_id
from mqtt_rest.mqtt import MQTTBroker, PROCESSING_DELAY

logger = logging.getLogger("uvicorn.error")
mqttclient = MQTTBroker()


class BaseSensor(BaseModel, abc.ABC):
    name: str
    unique_id: str = Field(serialization_alias="uniq_id")
    bulk_udpate: bool = Field(exclude=True)
    expire_after: Optional[int] = Field(default=None, serialization_alias="exp_aft")
    register_time: Optional[float] = Field(default=None, exclude=True)

    @abc.abstractmethod
    def get_config(self):
        pass

    def append_optional(self, dump: dict) -> dict:
        if self.bulk_udpate:
            dump["val_tpl"] = f"{{{{ value_json.{self.unique_id} }}}}"
        return dump

    @property
    def config_topic(self) -> str:
        component = "binary_sensor" if isinstance(self, BinarySensor) else "sensor"
        return f"homeassistant/{component}/{self.unique_id}/config"

    def get_state_topic(self, device_id: str) -> str:
        if self.bulk_udpate:
            return f"{CONFIG.app_name}/{device_id}/state"
        return f"{CONFIG.app_name}/{self.unique_id}/state"

    def wait_for_register(self):
        if self.register_time:
            remaining = PROCESSING_DELAY - (time.time() - self.register_time)
            if remaining > 0:
                time.sleep(remaining)

    def send_add(self, device_info: dict):
        device_id = device_info["ids"]
        config = self.get_config()
        config["stat_t"] = self.get_state_topic(device_id)
        config["dev"] = device_info
        # MQTT.publish(topic=self.config_topic, payload=config)
        self.wait_for_register()
        mqttclient.publish(topic=self.config_topic, payload=config)
        self.register_time = time.time()

    def send_remove(self):
        # MQTT.publish(topic=self.config_topic)
        self.wait_for_register()
        mqttclient.publish(topic=self.config_topic)
        self.register_time = time.time()


class BinarySensor(BaseSensor):
    value: bool = Field(exclude=True)
    device_class: Optional[str] = Field(default=None, serialization_alias="dev_cla")
    enabled_by_default: Optional[bool] = Field(default=None, serialization_alias="en")

    def get_config(self):
        return self.append_optional(self.model_dump(exclude_none=True, by_alias=True))


class ValueSensor(BaseSensor):
    value: Union[int, float] = Field(exclude=True)
    device_class: Optional[str] = Field(default=None, serialization_alias="dev_cla")
    unit: Optional[str] = Field(default=None, serialization_alias="unit_of_meas")
    state_class: Optional[str] = Field(default=None, serialization_alias="stat_cla")

    def get_config(self):
        return self.append_optional(self.model_dump(exclude_none=True, by_alias=True))


# class EnumSensor(BaseSensor):
#     value: str = Field(exclude=True)
#     options: list[str] = Field(serialization_alias="ops")

#     def get_config(self):
#         result = self.model_dump(exclude_none=True, by_alias=True)
#         result["dev_cla"] = "enum"
#         return result


class DiagnosticSensor(BaseSensor):
    value: str = Field(exclude=True)

    def get_config(self):
        result = self.append_optional(self.model_dump(exclude_none=True, by_alias=True))
        result["ent_cat"] = "diagnostic"
        return result


class Device:
    def __init__(
        self,
        name: str,
        model: Optional[str] = None,
        configuration_url: Optional[str] = None,
        via_device: Optional[str] = None,
        expire_after: Optional[int] = None,
    ):
        self.name = name
        self.model = model or "generic"
        self.id = get_unique_id(self.name + self.model)
        self.configuration_url = configuration_url
        self.via_device = via_device
        self.expire_after = expire_after
        self.discovered = False
        self.sensors = {}

    def get_config(self):
        dump = {
            "name": self.name,
            "id": self.id,
            "model": self.model,
            "discovered": self.discovered,
        }
        if self.configuration_url:
            dump["configuration_url"] = self.configuration_url
        if self.via_device:
            dump["via_device"] = self.via_device
        if self.expire_after:
            dump["expire_after"] = self.expire_after
        dump["sensors"] = {name: sensor.get_config() for name, sensor in self.sensors.items()}
        return dump

    def _create_sensor(self, name: str, value: Any, bulk: bool, **kwargs) -> BaseSensor:
        kwargs["name"] = name
        kwargs["unique_id"] = get_unique_id(self.id + name)
        kwargs["bulk_udpate"] = bulk
        kwargs["value"] = value
        if self.expire_after and "expire_after" not in kwargs:
            kwargs["expire_after"] = self.expire_after
        if isinstance(value, bool):
            return BinarySensor(**kwargs)
        if isinstance(value, (int, float)):
            return ValueSensor(**kwargs)
        kwargs["value"] = str(value)
        return DiagnosticSensor(**kwargs)

    def _get_device_info(self):
        if self.discovered:
            return {"ids": self.id}
        self.discovered = True
        info = {
            "name": self.name,
            "mdl": self.model,
            "mf": CONFIG.app_name,
            "sw": CONFIG.app_version,
            "ids": self.id,
        }
        if self.configuration_url:
            info["cu"] = self.configuration_url
        if self.via_device:
            info["via_device"] = self.via_device
        return info

    def remove_sensor(self, name: str):
        logger.info(f"Removing sensor {name} ({self.name})")
        if sensor := self.sensors.pop(name):
            sensor.send_remove()

    def get_sensor(self, name: str, value: Any, bulk: bool, **kwargs) -> BaseSensor:
        if sensor := self.sensors.get(name):
            if not isinstance(sensor.value, type(value)):
                logger.warning(
                    f"Sensor {name} ({self.name}) value type changed from {type(sensor.value)} to {type(value)}"
                )
                self.remove_sensor(name)
            elif sensor.bulk_udpate != bulk:
                logger.warning(
                    f"Sensor {name} ({self.name}) bulk update changed from {sensor.bulk_udpate} to {bulk}"
                )
                self.remove_sensor(name)
            else:
                return sensor

        logger.info(f"Registering sensor {name} ({self.name})")
        sensor = self._create_sensor(name, value, bulk, **kwargs)
        self.sensors[name] = sensor
        sensor.send_add(self._get_device_info())
        return sensor

    def update(self, name: str, value: Any):
        sensor = self.get_sensor(name, value, bulk=False)
        if isinstance(value, bool):
            value = "ON" if value else "OFF"
        # MQTT.publish(topic=sensor.get_state_topic(self.id), payload=value)
        sensor.wait_for_register()
        mqttclient.publish(topic=sensor.get_state_topic(self.id), payload=value)

    def bulk_update(self, data: Dict[str, Any]):
        values, topic, newest = {}, None, 0.0
        for name, value in data.items():
            sensor = self.get_sensor(name, value, bulk=True)
            newest = max(newest, sensor.register_time or 0)
            if not topic:
                topic = sensor.get_state_topic(self.id)
            if isinstance(value, bool):
                value = "ON" if value else "OFF"
            values[sensor.unique_id] = value
        if topic:
            # MQTT.publish(topic=topic, payload=values)
            remaining = PROCESSING_DELAY - (time.time() - newest)
            if remaining > 0:
                time.sleep(remaining)
            mqttclient.publish(topic=topic, payload=values)

    def bulk_remove(self, names: Optional[Iterable[str]] = None):
        if not names:
            names = list(self.sensors.keys())
        for name in names:
            self.remove_sensor(name)



import os
import importlib.metadata
from pathlib import Path
import toml

PYPROJECT_TOML_FILE = Path(__file__).parent.parent / "pyproject.toml"
if PYPROJECT_TOML_FILE.exists():
    PYPROJECT_TOML = toml.load(PYPROJECT_TOML_FILE)
    __app_name__ = PYPROJECT_TOML["tool"]["poetry"]["name"]
    __version__ = PYPROJECT_TOML["tool"]["poetry"]["version"]
else:
    __app_name__ = __package__ or __name__
    __version__ = importlib.metadata.version(__app_name__)

DEV_ENV_FILE = Path(__file__).parent.parent / "dev.env"
if DEV_ENV_FILE.exists():
    print(f"Loading environment variables from {DEV_ENV_FILE}")
    with open(DEV_ENV_FILE) as f:
        for line in f:
            key, value = line.strip().split("=")
            os.environ[key] = value



"""
Parse the output of the sensors command
and return a dictionary of the sensor values
"""

import re
from typing import Dict


def parse_sensors(output: str) -> Dict[str, float]:
    groups = output.split("\n\n")
    sensors = {}
    for group in groups:
        name = group.split("\n")[0]
        for line in group.split("\n"):
            match = re.match(r"^(.*?):\s+\+([0-9.]+)°C", line)
            if match:
                sensor, value = match.groups()
                sensors[f"{name} {sensor}"] = float(value)
    return sensors



import os
from typing import Optional, Callable, Any, Union
from pydantic import BaseModel, ConfigDict, computed_field
from mqtt_rest.utils import get_unique_id, get_internal_ip
from mqtt_rest import __app_name__, __version__


class ServerConfig(BaseModel):
    model_config = ConfigDict(frozen=True)

    app_name: str
    app_version: str
    ip: str
    port: int

    @computed_field
    def url(self) -> str:
        return f"http://{self.ip}:{self.port}"

    @computed_field
    def node_id(self) -> str:
        return get_unique_id(self.url)

    @computed_field
    def origin(self) -> dict:
        return {
            "name": self.app_name,
            "sw": self.app_version,
            "url": self.url,
        }


def get_env(name: str, defaut: Union[Any, Callable[[], Optional[Any]]]) -> Any:
    ip = os.environ.get(name)
    if not ip:
        if callable(defaut):
            ip = defaut()
        else:
            ip = defaut
    return ip


SERVER_CONFIG = ServerConfig(
    app_name=__app_name__,
    app_version=__version__,
    ip=get_env("SERVER_IP", get_internal_ip),
    port=get_env("SERVER_PORT", 9000),
)


class MQTTConfig(BaseModel):
    broker_ip: str
    broker_port: int
    user: str
    password: str
    enable_logger: bool


MQTT_CONFIG = MQTTConfig(
    broker_ip=get_env("BROKER_IP", "localhost"),
    broker_port=get_env("BROKER_PORT", 1883),
    user=get_env("MQTT_USER", "admin"),
    password=get_env("MQTT_PASS", ""),
    enable_logger=get_env("MQTT_LOG", "F").upper()[0] == "T",
)



import os
from fastapi.templating import Jinja2Templates

THIS_DIR = os.path.dirname(__file__)


def get_template(dir_path: str) -> Jinja2Templates:
    return Jinja2Templates(
        directory=dir_path,
        comment_start_string="{#=%",
        comment_end_string="%=#}",
    )


helper_templates = get_template(
    os.path.join(
        THIS_DIR,
        "_helper_commands",
    ),
)



import logging
from fastapi import APIRouter, Request
from fastapi.responses import PlainTextResponse
from mqtt_rest.plugins.helper import helper_templates
from mqtt_rest.plugins.sensors.parser import parse_sensors
from mqtt_rest.db import add_device, get_device, remove_device

PLUGIN_NAME = "sensors"
logger = logging.getLogger("uvicorn.error")
router = APIRouter(prefix=f"/{PLUGIN_NAME}")
last_report = {}


def get_device_name(name: str):
    name = name.replace(" ", "_")
    return f"{name} {PLUGIN_NAME}"


@router.get("/install")
async def get_install(request: Request):
    return helper_templates.TemplateResponse(
        "install.sh", {"request": request, "dependencies": "sensors"}
    )


@router.get("/manager")
async def get_manager(request: Request):
    return helper_templates.TemplateResponse(
        "single_job.sh", {"request": request, "data_command": "sensors"}
    )


@router.put("/submit/{name}")
async def submit(request: Request, name: str, expiry: int = 3600):
    global last_report
    device_name = get_device_name(name)
    last_report[device_name] = (await request.body()).decode()
    values = parse_sensors(last_report[device_name])
    device = get_device(device_name, create=False)
    if not device:
        configuration_url = str(request.url).replace("/submit/", "/report/")
        device = add_device(
            name=device_name,
            model="mqtt-plugin",
            configuration_url=configuration_url,
            expire_after=expiry,
        )
        for key, value in values.items():
            device.get_sensor(key, value, bulk=True, unit="°C")
        device.get_sensor("ip", "detecting...", bulk=False)
    device.bulk_update(values)
    device.update("ip", str(request.client.host))


@router.delete("/delete/{name}")
async def delete(name: str):
    global last_report
    device_name = get_device_name(name)
    last_report.pop(device_name)
    remove_device(device_name)


@router.get("/report/{name}", response_class=PlainTextResponse)
async def report(name: str):
    return last_report.get(get_device_name(name))



import time
from threading import Lock
from typing import Union
import logging
import json
import paho.mqtt.client as mqtt

from mqtt_rest.configs import MQTT_CONFIG, SERVER_CONFIG

logger = logging.getLogger("uvicorn.error")
PROCESSING_DELAY = 0.2


class SingletonMeta(type):
    """
    This is a thread-safe implementation of Singleton.
    see: https://refactoring.guru/design-patterns/singleton/python/example#example-1
    """

    _instances = {}
    _lock: Lock = Lock()

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]


class MQTTBroker(metaclass=SingletonMeta):
    client: mqtt.Client = None
    is_connected: bool = False

    def __init__(self) -> None:
        self.client = mqtt.Client(
            callback_api_version=mqtt.CallbackAPIVersion.VERSION2,
            client_id=SERVER_CONFIG.url,
            clean_session=1,
            protocol=mqtt.MQTTv311,
            transport="tcp",
            reconnect_on_failure=True,
        )
        self.client.on_connect = self._on_connect
        self.client.on_disconnect = self._on_disconnect
        self.client.username_pw_set(MQTT_CONFIG.user, MQTT_CONFIG.password)
        if MQTT_CONFIG.enable_logger:
            self.client.enable_logger(logger)

    def _on_connect(self, client, userdata, flags, rc, properties):
        if rc == 0:
            self.is_connected = True
            logger.info("Connected to MQTT Broker")
        else:
            logger.error(f"Failed to connect to MQTT Broker with code {rc}")

    def _on_disconnect(self, client, userdata, flags, rc, properties):
        self.is_connected = False
        logger.warning(f"Disconnected from MQTT Broker with code {rc}")

    def connect(self):
        if self.is_connected:
            logger.warning("Already connected to MQTT Broker (connect ignored)")
            return

        self.client.connect(
            host=MQTT_CONFIG.broker_ip,
            port=MQTT_CONFIG.broker_port,
            keepalive=3600,
        )
        self.client.loop_start()

        for i in range(10):
            if self.is_connected:
                break
            logger.info("Waiting for connection to MQTT Broker")
            time.sleep(1)
        else:
            logger.error("Failed to connect to MQTT Broker")
            self.client.loop_stop()
            return False
        return True

    def disconnect(self):
        if not self.is_connected:
            logger.warning("Not connected to MQTT Broker (disconnect ignored)")
            return
        time.sleep(PROCESSING_DELAY)  # wait for all messages to be sent
        self.client.disconnect()
        self.client.loop_stop()

    def publish(
        self, topic: str, payload: Union[str, dict] = "", retain: bool = False, qos: int = 1
    ):
        logger.debug(f"Publishing to {topic} (retain={retain}, qos={qos})")
        if isinstance(payload, dict):
            payload = json.dumps(payload)
        if not isinstance(payload, (int, float, str)):
            raise ValueError("Payload must be a string, int, or float")
        logger.debug(f"payload: {payload}")
        self.client.publish(topic, payload=payload, qos=qos, retain=retain)



import logging
from typing import Optional
from mqtt_rest.device import Device
from mqtt_rest.configs import SERVER_CONFIG as CONFIG

SOURCE_DEVICE = None
SOURCE_DEVICE_NAME = "devices"
all_devices = {}

logger = logging.getLogger("uvicorn.error")


def add_source_device():
    global SOURCE_DEVICE, SOURCE_DEVICE_NAME, all_devices
    SOURCE_DEVICE = Device(
        name=CONFIG.app_name,
        model=f"Server ({CONFIG.ip}:{CONFIG.port})",
        configuration_url=CONFIG.url + "/docs",
    )
    SOURCE_DEVICE.update(SOURCE_DEVICE_NAME, value=len(all_devices))


def remove_source_device():
    global SOURCE_DEVICE
    SOURCE_DEVICE.bulk_remove()
    SOURCE_DEVICE = None


def add_device(*args, **kwargs) -> Device:
    global SOURCE_DEVICE, SOURCE_DEVICE_NAME, all_devices
    kwargs["via_device"] = SOURCE_DEVICE.id
    logger.info(f"Adding device {kwargs['name']}")
    device = Device(*args, **kwargs)
    all_devices[device.name] = device
    SOURCE_DEVICE.update(SOURCE_DEVICE_NAME, value=len(all_devices))
    return device


def get_device(name: str, create: bool = True) -> Optional[Device]:
    global SOURCE_DEVICE, SOURCE_DEVICE_NAME, all_devices
    if name in all_devices:
        return all_devices[name]
    if not create:
        return None
    return add_device(name=name)


def remove_device(name: str):
    global SOURCE_DEVICE, SOURCE_DEVICE_NAME, all_devices
    device = all_devices.pop(name, None)
    if device:
        logger.info(f"Removing device {name}")
        device.bulk_remove()
        SOURCE_DEVICE.update(SOURCE_DEVICE_NAME, value=len(all_devices))



import os
import uvicorn
from mqtt_rest.server import app


def main():
    log_level = os.environ.get("SERVER_LOG_LEVEL", "info")
    uvicorn.run(app, host="0.0.0.0", port=9000, log_level=log_level)


if __name__ == "__main__":
    main()



from contextlib import asynccontextmanager
from typing import Dict, Union
from threading import Lock
from fastapi import FastAPI, HTTPException, Body
from mqtt_rest import db
from mqtt_rest.configs import SERVER_CONFIG as CONFIG
from mqtt_rest.plugins import all_plugins
from mqtt_rest.mqtt import MQTTBroker

mqttclient = MQTTBroker()


@asynccontextmanager
async def lifespan(app: FastAPI):
    print("Visit the API at", CONFIG.url + "/docs")
    if not mqttclient.connect():
        raise Exception("Failed to connect to MQTT Broker")
    db.add_source_device()
    yield
    print("Cleaning up devices")
    for device in db.all_devices.values():
        device.bulk_remove()
    db.remove_source_device()
    mqttclient.disconnect()


app = FastAPI(lifespan=lifespan)
lock = Lock()
for plugin in all_plugins.values():
    app.include_router(plugin, prefix="/api/v1/plugins")

# origins = "*"
# app.add_middleware(
#     CORSMiddleware,
#     allow_origins=origins,
#     allow_credentials=True,
#     allow_methods=["*"],
#     allow_headers=["*"],
# )


SensorValueType = Union[bool, int, float, str]
BulkSensorValuesType = Dict[str, SensorValueType]


@app.get("/api/v1/devices")
async def get_all_devices():
    return list(db.all_devices.keys())


@app.get("/api/v1/devices/{device_name}")
async def get_device(device_name: str):
    device = db.get_device(device_name, create=False)
    if not device:
        raise HTTPException(status_code=404, detail="Device not found")
    return device.get_config()


@app.put("/api/v1/devices/{device_name}")
async def put_device(device_name: str, values: BulkSensorValuesType = Body(...)):
    device = db.get_device(device_name)
    return device.bulk_update(values)


@app.delete("/api/v1/devices/{device_name}")
async def delete_device(device_name: str):
    db.remove_device(device_name)


@app.put("/api/v1/devices/{device_name}/sensors/{sensor_name}")
async def put_sensor(
    device_name: str, sensor_name: str, value: SensorValueType = Body(..., embed=True)
):
    device = db.get_device(device_name)
    return device.update(sensor_name, value)


@app.delete("/api/v1/devices/{device_name}/sensors/{sensor_name}")
async def delete_sensor(device_name: str, sensor_name: str):
    device = db.get_device(device_name, create=False)
    if not device:
        raise HTTPException(status_code=404, detail="Device not found")
    device.remove_sensor(sensor_name)


@app.get("/api/v1/plugins")
async def get_plugins():
    return list(all_plugins.keys())
